#!/usr/bin/env python
"""
Contains classes representing quantities which can be calculated from nested
sampling run.

Each estimator class should contain a mandatory member function returning the
value of the estimator for a nested sampling run:

    def __call__(self, ns_run, logw=None, simulate=False):
        ...

This allows logw to be provided if many estimators are being calculated from
the same run so logw is only calculated once. Otherwise logw is calculated from
the run if required.

They may also optionally contain a function giving its analytical value for
some given set of calculation settings (for use in checking results):

    def analytical(self, settings):
        ...

as well as helper functions.
Estimators should also contain class variables:

    name: str
        used for results tables.
    latex_name: str
        used for plotting results diagrams.

"""

import functools
import numpy as np
import scipy
import perfectns.maths_functions as mf
import nestcheck.ns_run_utils
import nestcheck.estimators


# Estimators
# ----------

class EstimatorBase(object):

    """Base class for estimators."""

    def __init__(self, func, **kwargs):
        """Set up estimator object, including making latex name and saving
        kwargs.

        Parameters
        ----------
        func: function
        kwargs: dict, optional
            Saved keyword arguments for function.
        """
        if kwargs:
            self.func = functools.partial(func, **kwargs)
        else:
            self.func = func
        self.latex_name = nestcheck.estimators.get_latex_name(func, **kwargs)

    def __call__(self, *args, **kwargs):
        """Returns estimator value for run."""
        return self.func(*args, **kwargs)


class LogZ(EstimatorBase):

    """Natural log of Bayesian evidence."""

    def __init__(self):
        """See EstimatorBase __init__ docstring."""
        EstimatorBase.__init__(self, nestcheck.estimators.logz)

    @staticmethod
    def analytical(settings):
        """Returns analytical value of estimator given settings."""
        return settings.logz_analytic()


class Z(EstimatorBase):

    """Bayesian evidence."""

    def __init__(self):
        """See EstimatorBase __init__ docstring."""
        EstimatorBase.__init__(self, nestcheck.estimators.evidence)

    @staticmethod
    def analytical(settings):
        """Returns analytical value of estimator given settings."""
        return np.exp(settings.logz_analytic())


class CountSamples(EstimatorBase):

    """Number of samples in run."""

    def __init__(self):
        """See EstimatorBase __init__ docstring."""
        EstimatorBase.__init__(self, nestcheck.estimators.count_samples)


class ParamMean(EstimatorBase):

    """
    Mean of a single parameter (single component of theta).
    By symmetry all parameters have the same distribution.
    """

    def __init__(self, param_ind=0):
        """See EstimatorBase __init__ docstring."""
        EstimatorBase.__init__(self, nestcheck.estimators.param_mean,
                               param_ind=param_ind)

    @staticmethod
    def analytical(settings):
        """Returns analytical value of estimator given settings."""
        return 0.

    @staticmethod
    def ftilde(logx, settings):
        """
        Helper function for finding the analytic value of the estimator.
        See "check_by_integrating" docstring for more details.

        ftilde(X) is mean of f(theta) on the iso-likelihood contour
        L(theta) = L(X).
        """
        return np.zeros(logx.shape)


class ParamCred(EstimatorBase):

    """
    One-tailed credible interval on the value of a single parameter (component
    of theta).
    By symmetry all parameters have the same distribution.
    """

    def __init__(self, probability, param_ind=0):
        """See EstimatorBase __init__ docstring."""
        self.probability = probability
        EstimatorBase.__init__(self, nestcheck.estimators.param_cred,
                               probability=probability, param_ind=param_ind)

    def analytical(self, settings):
        """Returns analytical value of estimator given settings."""
        if self.probability == 0.5:
            # by symmetry the median of any parameter given spherically
            # symmetric likelihoods and priors co-centred on zero is zero
            return 0
        else:
            assert type(settings.likelihood).__name__ == 'Gaussian', \
                "so far only set up for Gaussian likelihoods"
            assert type(settings.prior).__name__ in ['Gaussian',
                                                     'GaussianCached'], \
                "so far only set up for Gaussian priors"
            # the product of two Gaussians is another Gaussian with sigma:
            sigma = ((settings.likelihood.likelihood_scale ** -2) +
                     (settings.prior.prior_scale ** -2)) ** -0.5
            # find numCer of sigma from the mean by inverting the CDF of the
            # normal distribution.
            # CDF(x) = (1/2) + (1/2) * error_function(x / sqrt(2))
            zscore = (scipy.special.erfinv((self.probability * 2) - 1)
                      * np.sqrt(2))
            return zscore * sigma


class ParamSquaredMean(EstimatorBase):

    """
    Mean of the square of single parameter (second moment of its posterior
    distribution).
    By symmetry all parameters have the same distribution.
    """

    min_value = 0

    def __init__(self, param_ind=0):
        """See EstimatorBase __init__ docstring."""
        EstimatorBase.__init__(self, nestcheck.estimators.param_squared_mean,
                               param_ind=param_ind)

    @staticmethod
    def ftilde(logx, settings):
        """
        Helper function for finding the analytic value of the estimator.
        See "check_by_integrating" docstring for more details.

        ftilde(X) is mean of f(theta) on the iso-likelihood contour
        L(theta) = L(X).
        """
        # by symmetry at each (hyper)spherical iso-likelihood contour:
        r = settings.r_given_logx(logx)
        return r ** 2 / settings.n_dim

    def analytical(self, settings):
        """Returns analytical value of estimator given settings."""
        return check_by_integrating(self.ftilde, settings)


class RMean(EstimatorBase):

    """Mean of |theta| (the radial distance from the centre)."""

    min_value = 0

    def __init__(self, from_theta=False):
        """See EstimatorBase __init__ docstring."""
        EstimatorBase.__init__(self, nestcheck.estimators.r_mean)
        self.from_theta = from_theta

    def __call__(self, ns_run, logw=None, simulate=False):
        """
        Overwrite __call__ from nestcheck r_mean to allow use of perfectns'
        'r' key if from_theta=False, and to check all the dimensions have been
        sampled if from_theta=True.
        """
        if self.from_theta:
            # If run contains a dims_to_sample setting, check that samples from
            # every dimension are included
            assert (ns_run['settings']['dims_to_sample'] ==
                    ns_run['settings']['n_dim']), "Cannot work out radius!"
            return self.func(ns_run, logw=logw, simulate=simulate)
        else:
            if logw is None:
                logw = nestcheck.ns_run_utils.get_logw(
                    ns_run, simulate=simulate)
            w_relative = np.exp(logw - logw.max())
            r = ns_run['r']
            return np.sum(w_relative * r) / np.sum(w_relative)

    def analytical(self, settings):
        """Returns analytical value of estimator given settings."""
        return check_by_integrating(self.ftilde, settings)

    @staticmethod
    def ftilde(logx, settings):
        """
        Helper function for finding the analytic value of the estimator.
        See "check_by_integrating" docstring for more details.

        ftilde(X) is mean of f(theta) on the iso-likelihood contour
        L(theta) = L(X).
        """
        return settings.r_given_logx(logx)


class RCred(EstimatorBase):

    """One-tailed credible interval on the value of |theta|."""

    min_value = 0

    def __init__(self, probability, from_theta=False):
        """See EstimatorBase __init__ docstring."""
        self.probability = probability
        EstimatorBase.__init__(self, nestcheck.estimators.r_cred,
                               probability=probability)
        self.from_theta = from_theta

    def __call__(self, ns_run, logw=None, simulate=False):
        """Returns estimator value for run."""
        if self.from_theta:
            # If run contains a dims_to_sample setting, check that samples from
            # every dimension are included
            assert (ns_run['settings']['dims_to_sample'] ==
                    ns_run['settings']['n_dim']), "Cannot work out radius!"
            return self.func(ns_run, logw=logw, simulate=simulate,
                             probability=self.probability)
        else:
            if logw is None:
                logw = nestcheck.ns_run_utils.get_logw(
                    ns_run, simulate=simulate)
            # get sorted array of r values with their posterior weight
            wr = np.zeros((logw.shape[0], 2))
            wr[:, 0] = np.exp(logw - logw.max())
            wr[:, 1] = ns_run['r']
            wr = wr[np.argsort(wr[:, 1], axis=0)]
            # calculate cumulative distribution function (cdf)
            # Adjust by subtracting 0.5 * weight of first point to correct skew
            # - otherwise we need cdf=1 to return the last value but will
            # return the smallest value if cdf<the fractional weight of the
            # first point.
            # This should not much matter as typically points' relative weights
            # will be very small compared to self.probability or
            # 1-self.probability.
            cdf = np.cumsum(wr[:, 0]) - (wr[0, 0] / 2)
            cdf /= np.sum(wr[:, 0])
            # calculate cdf
            # linearly interpolate value
            return np.interp(self.probability, cdf, wr[:, 1])


# Functions for checking estimator results
# ----------------------------------------


def get_true_estimator_values(estimators, settings):
    """
    Calculates analytic values for estimators given the likelihood and
    prior in settings. If there is no method for calculating the values
    then np.nan is returned.

    Parameters
    ----------
    estimators: estimator object or list of estimator objects
    settings: PerfectNSSettings object

    Returns
    -------
    output: np.array of size (len(estimators),) if estimators is a list, float
        otherwise.
    """
    if isinstance(estimators, list):
        output = np.zeros(len(estimators))
        for i, est in enumerate(estimators):
            try:
                output[i] = est.analytical(settings)
            except (AttributeError, AssertionError):
                output[i] = np.nan
        return output
    else:
        try:
            return estimators.analytical(settings)
        except (AttributeError, AssertionError):
            return np.nan


def check_by_integrating(ftilde, settings):
    """
    Return the true value of the estimator using numerical
    integration.

    Chopin and Robert (2010) show that the expectation of some function
    f(theta) is given by the integral

        int L(X) X ftilde(X) dX / Z,

    where ftilde(X) is mean of f(theta) on the iso-likelihood contour
    L(theta) = L(X).

    Parameters
    ----------
    ftilde: function
    settings: PerfectNSSettings object

    Returns
    -------
    float
        The estimator's true value.
    """
    logx_terminate = mf.analytic_logx_terminate(settings)
    assert logx_terminate is not None, \
        'logx_terminate function not set up for current settings'
    result = scipy.integrate.quad(check_integrand, logx_terminate,
                                  0.0, args=(ftilde, settings))
    return result[0] / np.exp(settings.logz_analytic())


def check_integrand(logx, ftilde, settings):
    """Helper function to return integrand L(X) X ftilde(X) for checking
    estimator values by numerical integration.
    Note that the integral must be normalised by multiplying by a factor of
    (1/Z).

    Parameters
    ----------
    logx: 1d numpy array
        Values on which to evaluate integrand.
    ftilde: function
        See check_by_integrating docstring for more details.
    settings: PerfectNSSettings object

    Returns
    -------
    1d numpy array
        Integrand evaluated at input logx coordinates.
    """
    # returns L(X) X ftilde(X) for integrating dlogx
    return (np.exp(settings.logl_given_logx(logx) + logx)
            * ftilde(logx, settings))



#!/usr/bin/env python
"""
Classes representing spherically symmetric likelihoods.

Each likelihood class must contain a member function giving the log likelihood
as a function of the radial coordinate r = |theta| and the number of dimensions

    def logl_given_r(self, r, n_dim):
        ...

The number of dimensions is not stored in this class but in the
PerfectNSSettings object to ensure it is the same for the likelihood and the
prior.

Likelihood classes may also optionally contain the inverse function

    def r_given_logl(self, logl, n_dim):
        ...

(although this is not needed to generate nested sampling runs). Another
optional function is the analytic value of the log evidence for some
given prior and dimension, which is useful for checking results:

    def logz_analytic(self, prior, n_dim):
        ...
"""

import numpy as np
import perfectns.maths_functions as mf


class Gaussian(object):

    """Spherically symmetric Gaussian likelihood."""

    def __init__(self, likelihood_scale=1.0):
        """Store Gaussian likelihood's sigma."""
        self.likelihood_scale = likelihood_scale

    def logl_given_r(self, r, n_dim):
        """
        Get loglikelihood values for input radial coordinates.

        Parameters
        ----------
        r: float or numpy array
            Radial coordinates.
        n_dim: int
            Number of dimensions.

        Returns
        -------
        logl: same type and size as r
            Loglikelihood values.
        """
        return mf.log_gaussian_given_r(r, self.likelihood_scale, n_dim)

    # Optional functions

    def r_given_logl(self, logl, n_dim):
        """
        Get the radial coordinates corresponding to the input loglikelihood
        values.

        Parameters
        ----------
        logl: float or numpy array
            Loglikelihood values.
        n_dim: int
            Number of dimensions.

        Returns
        -------
        r: same type and size as logl
            Radial coordinates.
        """
        return mf.r_given_log_gaussian(logl, self.likelihood_scale, n_dim)

    def logz_analytic(self, prior, n_dim):
        """
        Returns analytic value of the log evidence for the input prior and
        dimension if it is available.

        If not set up for this prior an AssertionError is thrown (this is
        caught in functions which check analytical values where they are
        available).

        Parameters
        ----------
        prior: object
        n_dim: int
            Number of dimensions.

        Returns
        -------
        float
            Analytic value of log Z for this likelihood given the prior and
            number of dimensions.
        """
        assert type(prior).__name__ in ['Uniform', 'Gaussian'], \
            ('No logz_analytic set up for ' + type(self).__name__ +
             ' likelihoods and ' + type(prior).__name__ + ' priors.')
        if type(prior).__name__ == 'Uniform':
            # logZ = -log volume of uniform prior + correction for truncation
            logvol = mf.nsphere_logvol(n_dim, radius=prior.prior_scale)
            # To find how much of the likelihood's mass lies within the uniform
            # prior we can reuse the gaussian_logx_given_r function which is
            # used for Gaussian priors.
            return -logvol + mf.gaussian_logx_given_r(prior.prior_scale,
                                                      self.likelihood_scale,
                                                      n_dim)
        else:
            assert type(prior).__name__ == 'Gaussian'
            # See 'Products and convolutions of Gaussian probability density
            # functions' (P Bromiley, 2003) page 3 for a derivation of this
            # result
            return (-n_dim / 2.) * np.log(2 * np.pi *
                                          (self.likelihood_scale ** 2 +
                                           prior.prior_scale ** 2))


class ExpPower(object):

    """
    Spherically symmetric exponential power likelihood.
    When power=1, this is the same as the Gaussian likelihood.
    """

    def __init__(self, likelihood_scale=1, power=2):
        """Save the likelihood scale and power."""
        self.likelihood_scale = likelihood_scale
        self.power = power

    def logl_given_r(self, r, n_dim):
        """
        Get loglikelihood values for input radial coordinates.

        Parameters
        ----------
        r: float or numpy array
            Radial coordinates.
        n_dim: int
            Number of dimensions.

        Returns
        -------
        logl: same type and size as r
            Loglikelihood values.
        """
        return mf.log_exp_power_given_r(r, self.likelihood_scale,
                                        n_dim, b=self.power)

    # Optional functions

    def r_given_logl(self, logl, n_dim):
        """
        Get the radial coordinates corresponding to the input loglikelihood
        values.

        Parameters
        ----------
        logl: float or numpy array
            Loglikelihood values.
        n_dim: int
            Number of dimensions.

        Returns
        -------
        r: same type and size as logl
            Radial coordinates.
        """
        return mf.r_given_log_exp_power(logl, self.likelihood_scale,
                                        n_dim, b=self.power)


class Cauchy(object):

    """Spherically symmetric Cauchy likelihood."""

    def __init__(self, likelihood_scale=1):
        """Save the likelihood scale."""
        self.likelihood_scale = likelihood_scale

    def logl_given_r(self, r, n_dim):
        """
        Get loglikelihood values for input radial coordinates.

        Parameters
        ----------
        r: float or numpy array
            Radial coordinates.
        n_dim: int
            Number of dimensions.

        Returns
        -------
        logl: same type and size as r
            Loglikelihood values.
        """
        return mf.log_cauchy_given_r(r, self.likelihood_scale, n_dim)

    # Optional functions

    def r_given_logl(self, logl, n_dim):
        """
        Get the radial coordinates corresponding to the input loglikelihood
        values.

        Parameters
        ----------
        logl: float or numpy array
            Loglikelihood values.
        n_dim: int
            Number of dimensions.

        Returns
        -------
        r: same type and size as logl
            Radial coordinates.
        """
        return mf.r_given_log_cauchy(logl, self.likelihood_scale, n_dim)



#!/usr/bin/env python
"""
Functions which perform standard and dynamic nested sampling runs and generate
samples for use in evidence calculations and parameter estimation.

Nested sampling runs are stored a format compatible with the nestcheck
package.
"""

import warnings
import copy
import numpy as np
import scipy.special
import nestcheck.ns_run_utils
import nestcheck.parallel_utils as pu
import nestcheck.io_utils as iou
import perfectns.maths_functions as mf


def generate_ns_run(settings, random_seed=None):
    """
    Performs perfect nested sampling calculation and returns a nested sampling
    run in the form of a dictionary.

    This function is just a wrapper around the
    generate_standard_run (performs standard nested sampling) and
    generate_dynamic_run (performs dynamic nested sampling) which are chosen
    depending on the input settings.

    Parameters
    ----------
    settings: PerfectNSSettings object
    random_seed: None, bool or int, optional
        Set numpy random seed. Default is to use None (so a random seed is
        chosen from the computer's internal state) to ensure reliable results
        when multiprocessing. Can set to an integer or to False to not edit the
        seed.

    Returns
    -------
    dict
        Nested sampling run dictionary containing information about the run's
        posterior samples and a record of the settings used. These are as
        separate arrays giving values for points in order of increasing
        likelihood.
        Keys:
            'settings': dict recording settings used.
            'logl': 1d array of log likelihoods.
            'r': 1d array of radial coordinates.
            'logx': 1d array of logx coordinates.
            'theta': 2d array, each row is sample coordinate. The number of
                     co-ordinates saved is controlled by
                     settings.dims_to_sample.
            'nlive_array': 1d array of the local number of live points at each
                           sample.
            'thread_min_max': 2d array containing likelihoods at which each
                              nested sampling thread begins and ends.
            'thread_labels': 1d array listing the threads each sample belongs
                              to.
    """
    if random_seed is not False:
        np.random.seed(random_seed)
    if settings.dynamic_goal is None:
        run = generate_standard_run(settings)
    else:
        run = generate_dynamic_run(settings)
    run['random_seed'] = random_seed
    return run


def get_run_data(settings, n_repeat, **kwargs):
    """
    Tests if runs with the specified settings are already cached. If not
    the runs are generated and saved.

    Parameters
    ----------
    settings: PerfectNSSettings object
    n_repeat: int
        Number of nested sampling runs to generate.
    parallel: bool, optional
        Should runs be generated in parallel?
    max_workers: int or None, optional
        Number of processes.
        If max_workers is None then concurrent.futures.ProcessPoolExecutor
        defaults to using the number of processors of the machine.
        N.B. If max_workers=None and running on supercomputer clusters with
        multiple nodes, this may default to the number of processors on a
        single node and therefore there will be no speedup from multiple
        nodes (must specify manually in this case).
    load: bool, optional
        Should previously saved runs be loaded? If False, new runs are
        generated.
    save: bool, optional
        Should any new runs generated be saved?
    cache_dir: str, optional
        Directory for caching
    overwrite_existing: bool, optional
        if a file exists already but we generate new run data, should we
        overwrite the existing file when saved?
    check_loaded_settings: bool, optional
        if we load a cached file, should we check if the loaded file's settings
        match the current settings (and generate fresh runs if they do not)?
    random_seeds: list, optional
        random_seed arguments for each call of generate_ns_run.

    Returns
    -------
    run_list
        list of n_repeat nested sampling runs.
    """
    parallel = kwargs.pop('parallel', True)
    max_workers = kwargs.pop('max_workers', None)
    load = kwargs.pop('load', True)
    save = kwargs.pop('save', True)
    cache_dir = kwargs.pop('cache_dir', 'cache')
    overwrite_existing = kwargs.pop('overwrite_existing', True)
    check_loaded_settings = kwargs.pop('check_loaded_settings', True)
    random_seeds = kwargs.pop('random_seeds', [None] * n_repeat)
    assert len(random_seeds) == n_repeat
    if kwargs:
        raise TypeError('Unexpected **kwargs: {0}'.format(kwargs))
    save_name = cache_dir + '/' + settings.save_name()
    save_name += '_' + str(n_repeat) + 'reps'
    if load:
        try:
            data = iou.pickle_load(save_name)
            if check_loaded_settings:
                # Assume all runs in the loaded list have the same settings, in
                # which case we only need check the first one.
                loaded = copy.deepcopy(data[0]['settings'])
                current = copy.deepcopy(settings.get_settings_dict())
                # If runs are standard nested sampling there is no need to
                # check settings which only affect dynamic ns match
                if loaded['dynamic_goal'] is None and (current['dynamic_goal']
                                                       is None):
                    for key in ['dynamic_goal', 'n_samples_max', 'ninit',
                                'nbatch', 'dynamic_fraction',
                                'tuned_dynamic_p']:
                        del loaded[key]
                        del current[key]
                if loaded != current:
                    #  remove shared keys and only print differences
                    rm = [k for k in set(loaded.keys()) & set(current.keys())
                          if loaded[k] == current[k]]
                    loaded_diff = {k: v for k, v in loaded.items() if k
                                   not in rm}
                    current_diff = {k: v for k, v in current.items() if k
                                    not in rm}
                    msg = (('Loaded settings != current settings. Differences '
                            'are: {0} != = {1}. Generating new runs instead.')
                           .format(loaded_diff, current_diff))
                    warnings.warn(msg, UserWarning)
                    del data
                    load = False
        except (OSError, EOFError) as exception:
            print(('Loading {0} failed due to {1}'
                   .format(save_name, type(exception).__name__)
                   + ' - try generating new runs instead.'))
            load = False
            overwrite_existing = True
    if not load:
        # Must check cache is up to date before parallel_apply or each process
        # will have to update the cache seperately
        if type(settings.prior).__name__ == 'GaussianCached':
            settings.prior.check_cache(settings.n_dim)
        data = pu.parallel_apply(generate_ns_run, random_seeds,
                                 func_pre_args=(settings,),
                                 max_workers=max_workers,
                                 parallel=parallel)
        if save:
            iou.pickle_save(data, save_name,
                            overwrite_existing=overwrite_existing)
    return data


def generate_standard_run(settings, is_dynamic_initial_run=False):
    """
    Performs standard nested sampling using the likelihood and prior specified
    in settings.

    The run terminates when the estimated posterior mass contained in the live
    points is less than settings.termination_fraction. The evidence in the
    remaining live points is estimated as

        Z_{live} = average likelihood of live points * prior volume remaining

    Parameters
    ----------
    settings: PerfectNSSettings object
    is_dynamic_initial_run: bool, optional
        Set to True if this is the initial exploratory run in dynamic nested
        sampling.

    Returns
    -------
    run: dict
        Nested sampling run dictionary containing information about the run's
        posterior samples and a record of the settings used. See docstring for
        generate_ns_run for more details.
    """
    if is_dynamic_initial_run:
        nlive_const = settings.ninit
    else:
        nlive_const = settings.nlive_const
    # Sample live points as a 2-dimensional array with columns:
    # [loglikelihood, radial coordinate, logx coordinate, thread label]
    live_points = np.zeros((nlive_const, 4))
    # Thread labels are 1 to nlive_const
    live_points[:, 3] = np.arange(nlive_const)
    live_points[:, 2] = np.log(np.random.random(live_points.shape[0]))
    live_points[:, 1] = settings.r_given_logx(live_points[:, 2])
    live_points[:, 0] = settings.logl_given_r(live_points[:, 1])
    # termination condition variables
    logx_i = 0.0
    logz_dead = -np.inf
    logz_live = (scipy.special.logsumexp(live_points[:, 0]) + logx_i -
                 np.log(nlive_const))
    # Calculate factor for trapezium rule of geometric series
    shrinkage = np.exp(-1.0 / nlive_const)
    logtrapz = np.log(0.5 * ((shrinkage ** -1) - shrinkage))
    # start the array of dead points
    dead_points_list = []
    while logz_live - np.log(settings.termination_fraction) > logz_dead:
        # add to dead points
        ind = np.where(live_points[:, 0] == live_points[:, 0].min())[0][0]
        dead_points_list.append(copy.deepcopy(live_points[ind, :]))
        # update dead evidence estimates
        logx_i += -1.0 / nlive_const
        logz_dead = scipy.special.logsumexp((logz_dead, live_points[ind, 0] +
                                             logtrapz + logx_i))
        # add new point
        live_points[ind, 2] += np.log(np.random.random())
        live_points[ind, 1] = settings.r_given_logx(live_points[ind, 2])
        live_points[ind, 0] = settings.logl_given_r(live_points[ind, 1])
        logz_live = (scipy.special.logsumexp(live_points[:, 0]) + logx_i -
                     np.log(nlive_const))
    points = np.array(dead_points_list)
    # add remaining live points (sorted by increasing likelihood)
    points = np.vstack((points, live_points[np.argsort(live_points[:, 0])]))
    # Create a dictionary representing the nested sampling run
    run = {'settings': settings.get_settings_dict(),
           'logl': points[:, 0],
           'r': points[:, 1],
           'logx': points[:, 2],
           'thread_labels': points[:, 3].astype(int)}
    # add array of parameter values sampled from the hyperspheres corresponding
    # to the radial coordinate of each point.
    run['theta'] = mf.sample_nsphere_shells(run['r'], settings.n_dim,
                                            settings.dims_to_sample)
    # Add an array of the local number of live points - this equals nlive_const
    # until the run terminates, at which point it reduces by 1 as each thread
    # ends.
    run['nlive_array'] = np.zeros(run['logl'].shape[0]) + nlive_const
    for i in range(1, nlive_const):
        run['nlive_array'][-i] = i
    # Get array of data on threads' beginnings and ends. Each starts by
    # sampling the whole prior and ends on one of the final live points.
    run['thread_min_max'] = np.zeros((nlive_const, 2))
    run['thread_min_max'][:, 0] = -np.inf
    run['thread_min_max'][:, 1] = live_points[:, 0]
    return run


# Make dynamic ns run:
# --------------------


def generate_dynamic_run(settings):
    """
    Generate a dynamic nested sampling run.
    For details of the dynamic nested sampling algorithm, see 'Dynamic nested
    sampling: an improved algorithm for nested sampling parameter estimation
    and evidence calculation' (Higson et al., 2017).

    The run terminates when the number of samples reaches some limit
    settings.n_samples_max. If this is not set, the function will estimate the
    number of samples that a standard nested sampling run with
    settings.nlive_const would use from the number of samples in the initial
    exploratory run.

    Parameters
    ----------
    settings: PerfectNSSettings object
        settings.dynamic_goal controls whether the algorithm aims to increase
        parameter estimation accuracy (dynamic_goal=1), evidence accuracy
        (dynamic_goal=0) or places some weight on both.

    Returns
    -------
    dict
        Nested sampling run dictionary containing information about the run's
        posterior samples and a record of the settings used. See docstring for
        generate_ns_run for more details.
    """
    assert 1 >= settings.dynamic_goal >= 0, 'dynamic_goal = ' + \
        str(settings.dynamic_goal) + ' should be between 0 and 1'
    # Step 1: initial exploratory standard ns run with ninit live points
    # ------------------------------------------------------------------
    standard_run = generate_standard_run(settings, is_dynamic_initial_run=True)
    # create samples array with columns:
    # [logl, r, logx, thread label, change in nlive, params]
    samples = samples_array_given_run(standard_run)
    thread_min_max = standard_run['thread_min_max']
    n_samples = samples.shape[0]
    n_samples_max = copy.deepcopy(settings.n_samples_max)
    if n_samples_max is None:
        # estimate number of likelihood calls available
        n_samples_max = n_samples * (settings.nlive_const / settings.ninit)
    # Step 2: add samples wherever they are most useful
    # -------------------------------------------------
    while n_samples < n_samples_max:
        importance = point_importance(samples, thread_min_max, settings)
        logl_min_max, logx_min_max = min_max_importance(importance,
                                                        samples,
                                                        settings)
        for _ in range(settings.nbatch):
            # make new thread
            thread_label = thread_min_max.shape[0]
            thread = generate_single_thread(settings,
                                            logx_min_max[1],
                                            thread_label,
                                            logx_start=logx_min_max[0],
                                            keep_final_point=True)
            # update run
            if logl_min_max[0] != -np.inf:
                start_ind = np.where(samples[:, 0] == logl_min_max[0])[0]
                # check there is exactly one point with the likelihood at which
                # the new thread starts, and note that nlive increases by 1
                assert start_ind.shape == (1,)
                samples[start_ind, 4] += 1
            samples = np.vstack((samples, thread))
            lmm = np.asarray([logl_min_max[0], thread[-1, 0]])
            thread_min_max = np.vstack((thread_min_max, lmm))
        # sort array and update n_samples in preparation for the next iteration
        samples = samples[np.argsort(samples[:, 0])]
        n_samples = samples.shape[0]
    # To compute nlive from the changes in nlive at each step, first find nlive
    # for the first point (= the number of threads which sample from the entire
    # prior)
    run = dict_given_samples_array(samples, thread_min_max)
    run['settings'] = settings.get_settings_dict()
    return run


# Dynamic NS helper functions
# ------------------------------


def generate_thread_logx(logx_end, logx_start=0, keep_final_point=True):
    """
    Generate logx co-ordinates of a new nested sampling thread (single live
    point run).

    Parameters
    ----------
    logx_end: float
        Logx value at which run terminates.
    logx_start: float, optional.
        Logx value at which run starts. 0 corresponds to sampling from the
        whole prior.
    keep_final_point: bool, optional
        If False, the final point with logx less than logx_end is removed.

    Returns
    -------
    logx_list: list of floats
    """
    logx_list = [logx_start + np.log(np.random.random())]
    while logx_list[-1] > logx_end:
        logx_list.append(logx_list[-1] + np.log(np.random.random()))
    if not keep_final_point:
        del logx_list[-1]  # remove point which violates termination condition
    return logx_list


def generate_single_thread(settings, logx_end, thread_label, logx_start=0,
                           keep_final_point=True):
    """
    Generates a samples array for a thread (single live point run) between
    logx_start and logx_end.
    Settings argument specifies how the calculation is done.

    Parameters
    ----------
    settings: PerfectNSSettings object
    logx_end: float
    thread_label: int
        Index labelling the thread.
    logx_start: float, optional
    keep_final_point: bool, optional
        See generate_thread_logx docstring.
    """
    assert logx_start > logx_end, 'generate_single_thread: logx_start=' + \
        str(logx_start) + ' <= logx_end=' + str(logx_end)
    logx_list = generate_thread_logx(logx_end, logx_start=logx_start,
                                     keep_final_point=keep_final_point)
    if not logx_list:  # PEP8 method for testing if sequence is empty
        return None
    else:
        lrxtn = np.zeros((len(logx_list), 5))
        lrxtn[:, 3] = thread_label
        lrxtn[:, 2] = np.asarray(logx_list)
        lrxtn[:, 1] = settings.r_given_logx(lrxtn[:, 2])
        lrxtn[:, 0] = settings.logl_given_r(lrxtn[:, 1])
        # Check there are non nans mapping logx to r
        assert np.all(~np.isnan(lrxtn[:, 1])), (
            'nans in r_given_logx(logx)=' + str(lrxtn[:, 1]) +
            '\nrows with nans are ' +
            str(lrxtn[np.where(np.isnan(lrxtn[:, 1]))[0], :]) +
            '\nperhaps the prior is numerically unstable?')
        # set change in nlive to -1 where thread ends (zero elsewhere)
        lrxtn[-1, 4] = -1
        theta = mf.sample_nsphere_shells(lrxtn[:, 1],
                                         settings.n_dim,
                                         settings.dims_to_sample)
        return np.hstack([lrxtn, theta])


def point_importance(samples, thread_min_max, settings, simulate=False):
    """
    Calculate the relative importance of each point for use in the dynamic
    nested sampling algorithm.

    For more details see 'Dynamic nested sampling: an improved algorithm for
    nested sampling parameter estimation and evidence calculation' (Higson et
    al., 2017).

    Parameters
    ----------
    samples: 2d numpy array
        See dict_given_samples_arrry docstring for details of columns.
    thread_min_max: 2d numpy array
        First column is starting logl of each thread and second column is
        ending logl.
    settings: PerfectNSSettings object
    simulate: bool, optional
        Passed to nestcheck.ns_run_utils.get_logw.

    Returns
    -------
    importance: 1d numpy array
        Relative point importances of the rows of the input samples array.
        Normalised so the biggest value in the array is equal to 1.
    """
    run_dict = dict_given_samples_array(samples, thread_min_max)
    logw = nestcheck.ns_run_utils.get_logw(run_dict, simulate=simulate)
    w_relative = np.exp(logw - logw.max())
    if settings.dynamic_goal == 0:
        return z_importance(w_relative, run_dict['nlive_array'])
    elif settings.dynamic_goal == 1:
        return p_importance(run_dict['theta'], w_relative,
                            tuned_dynamic_p=settings.tuned_dynamic_p)
    else:
        imp_z = z_importance(w_relative, run_dict['nlive_array'])
        imp_p = p_importance(run_dict['theta'], w_relative,
                             tuned_dynamic_p=settings.tuned_dynamic_p)
        importance = (imp_z / np.sum(imp_z)) * (1.0 - settings.dynamic_goal)
        importance += (imp_p / np.sum(imp_p)) * settings.dynamic_goal
        return importance / importance.max()


def z_importance(w_relative, nlive):
    """
    Calculate the relative importance of each point for evidence calculation.

    For more details see 'Dynamic nested sampling: an improved algorithm for
    nested sampling parameter estimation and evidence calculation'
    (Higson et al., 2017).

    Parameters
    ----------
    w_relative: 1d numpy array
        Relative point weights.
    nlive: 1d numpu array
        Number of live points.

    Returns
    -------
    importance: 1d numpy array
        Relative point importances.
        Normalised so the biggest value in the array is equal to 1.
    """
    importance = np.cumsum(w_relative)
    importance = importance.max() - importance
    importance *= 1.0 / nlive
    return importance / importance.max()


def p_importance(theta, w_relative, tuned_dynamic_p=False,
                 tuning_type='theta1'):
    """
    Calculate the relative importance of each point for parameter estimation.

    For more details see 'Dynamic nested sampling: an improved algorithm for
    nested sampling parameter estimation and evidence calculation' (Higson et
    al., 2017).

    Parameters
    ----------
    theta: 2d numpy array
        Each row gives parameter values of samples.
    w_relative: 1d numpy array
        Relative point weights.
    tuned_dynamic_p: bool, optional
        Whether or not to tune for a specific parameter estimation problem.
        See the dynamic nested sampling paper for more details.
    tuning_type: str, optional
        Which parameter estimation problem to tune for. Only used if
        tuned_dynamic_p is True. So far only set up to tune for the mean of the
        first parameter.

    Returns
    -------
    importance: 1d numpy array
        Relative point importances.
        Normalised so the biggest value in the array is equal to 1.
    """
    if tuned_dynamic_p is False:
        return w_relative / w_relative.max()
    else:
        assert tuning_type == 'theta1', 'so far only set up for theta1'
        if tuning_type == 'theta1':
            ftheta = theta[:, 0]
        # calculate importance in proportion to difference between f values and
        # the calculation mean.
        fabs = np.absolute(ftheta - (np.sum(ftheta * w_relative) /
                                     np.sum(w_relative)))
        importance = fabs * w_relative
        return importance / importance.max()


def min_max_importance(importance, samples, settings):
    """
    Find the logl and logx values at which to start and end additional dynamic
    nested sampling threads.

    Parameters
    ----------
    importance: 1d numpy array
        Relative importances of samples.
    samples: 2d numpy array
        See dict_given_samples_arrry docstring for details of columns.
    settings: PerfectNSSettings object

    Returns
    -------
    list of two floats
        Contains logl_min and logl_max defining the start and end of the region
        from which new points should be sampled.
    list of two floats
        Logx values corresponding to logl_min and logl_max.
    """
    assert settings.dynamic_fraction > 0. and settings.dynamic_fraction < 1., \
        'min_max_importance: settings.dynamic_fraction = ' + \
        str(settings.dynamic_fraction) + ' must be in [0, 1]'
    # where to start the additional threads:
    high_importance_inds = np.where(importance > settings.dynamic_fraction)[0]
    if high_importance_inds[0] == 0:  # start from sampling the whole prior
        logl_min = -np.inf
        logx_min = 0
    else:
        logl_min = samples[:, 0][high_importance_inds[0] - 1]
        # Use lookup to avoid recalculating the logx values (otherwise there
        # may be float comparison errors).
        ind = np.where(samples[:, 0] == logl_min)[0]
        assert ind.shape == (1,), \
            'Should be one unique match for logl=logl_min=' + str(logl_min) + \
            '. Instead we have matches at indexes ' + str(ind) + \
            ' of the samples array (shape ' + str(samples.shape) + ')'
        logx_min = samples[ind[0], 2]
    # where to end the additional threads:
    if high_importance_inds[-1] == samples[:, 0].shape[0] - 1:
        logl_max = samples[-1, 0]
        logx_max = samples[-1, 2]
    else:
        logl_max = samples[:, 0][(high_importance_inds[-1] + 1)]
        # Use lookup to avoid recalculating the logx values (otherwise there
        # may be float comparison errors).
        ind = np.where(samples[:, 0] == logl_max)[0]
        assert ind.shape == (1,), \
            'Should be one unique match for logl=logl_max=' + str(logl_max) + \
            '.\n Instead we have matches at indexes ' + str(ind) + \
            ' of the samples array (shape ' + str(samples.shape) + ')'
        logx_max = samples[ind[0], 2]
    return [logl_min, logl_max], [logx_min, logx_max]


def samples_array_given_run(ns_run):
    """
    Converts information on samples in a nested sampling run dictionary into a
    numpy array representation. This allows fast addition of more samples and
    recalculation of nlive.

    Parameters
    ----------
    ns_run: dict
        Nested sampling run dictionary.
        Contains keys: 'logl', 'r', 'logx', 'thread_label', 'nlive_array',
        'theta'

    Returns
    -------
    samples: numpy array
        Numpy array containing columns
        [logl, r, logx, thread label, change in nlive at sample, (thetas)]
        with each row representing a single sample.
    """
    samples = np.zeros((ns_run['logl'].shape[0], 5 + ns_run['theta'].shape[1]))
    samples[:, 0] = ns_run['logl']
    samples[:, 1] = ns_run['r']
    samples[:, 2] = ns_run['logx']
    samples[:, 3] = ns_run['thread_labels'].astype(int)
    # Calculate 'change in nlive' after each step
    samples[:-1, 4] = np.diff(ns_run['nlive_array'])
    samples[-1, 4] = -1  # nlive drops to zero after final point
    samples[:, 5:] = ns_run['theta']
    return samples


def dict_given_samples_array(samples, thread_min_max):
    """
    Converts an array of information about samples back into a dictionary.

    Parameters
    ----------
    samples: numpy array
        Numpy array containing columns
        [logl, r, logx, thread label, change in nlive at sample, (thetas)]
        with each row representing a single sample.
    thread_min_max': numpy array, optional
        2d array with a row for each thread containing the likelihoods at which
        it begins and ends.
        Needed to calculate nlive_array (otherwise this is set to None).

    Returns
    -------
    ns_run: dict
        Nested sampling run dictionary corresponding to the samples array.
        Contains keys: 'logl', 'r', 'logx', 'thread_label', 'nlive_array',
        'theta'
        N.B. this does not contain a record of the run's settings.
    """
    nlive_0 = (thread_min_max[:, 0] == -np.inf).sum()
    nlive_array = np.zeros(samples.shape[0]) + nlive_0
    nlive_array[1:] += np.cumsum(samples[:-1, 4])
    assert nlive_array.min() > 0, (
        'nlive contains 0s or negative values!' +
        '\nnlive_array = ' + str(nlive_array) +
        '\nfinal row of samples arr = ' + str(samples[-4:, :]) +
        '\nthread_min_max = ' + str(thread_min_max))
    assert nlive_array[-1] == 1, (
        'final point in nlive_array != 1!' +
        '\nnlive_array = ' + str(nlive_array) +
        '\nfinal row of samples arr = ' + str(samples[-4:, :]) +
        '\nthread_min_max = ' + str(thread_min_max))
    ns_run = {'logl': samples[:, 0],
              'r': samples[:, 1],
              'logx': samples[:, 2],
              'thread_labels': samples[:, 3].astype(int),
              'nlive_array': nlive_array,
              'thread_min_max': thread_min_max,
              'theta': samples[:, 5:]}
    return ns_run



#!/usr/bin/env python
"""
Contains helper functions for the 'gaussian_cached' prior.

This is needed as the 'gaussian' prior suffers from overflow errors and
numerical instability for very low values of X, which are reached in
high dimensional problems.
"""

import warnings
import numpy as np
import nestcheck.io_utils as iou
import perfectns.maths_functions as mf


def interp_r_logx_dict(n_dim, prior_scale, **kwargs):
    """
    Generate a dictionary containing arrays of logx and r values for use in
    interpolation, as well as a record of the settings used.

    Parameters
    ----------
    n_dim: int
        Number of dimensions.
    prior_scale: float
        Gaussian prior's standard deviation.
    logx_min: float
        Values for interpolation are generated between logx_min and logx_max.
    save_dict: bool, optional
        Whether or not to cache the output dictionary.
    cache_dir: str, optional
        Directory in which to cache output if save_dict is True.
    interp_density: float
        Number of points to include in interpolation arrays per unit of logx
        (the number of points is cast to int so this can be a fraction).
    logx_max: float, optional
        Values for interpolation are generated between logx_min and logx_max.

    Returns
    -------
    interp_dict: dict
    """
    logx_min = kwargs.pop('logx_min')  # no default, must specify
    save_dict = kwargs.pop('save_dict', True)
    cache_dir = kwargs.pop('cache_dir', 'cache')
    interp_density = kwargs.pop('interp_density')  # no default, must specify
    if n_dim > 1000 and logx_min >= -4500:
        warnings.warn(
            ('Interp_r_logx_dict: WARNING: n_dim=' + str(n_dim) + ': '
             'for very high dimensions, depending on the likelihood, you may '
             'need to lower logx_min=' + str(logx_min)), UserWarning)
    if n_dim < 50:
        warnings.warn(
            ('Interp_r_logx_dict: WARNING: n_dim=' + str(n_dim) + ': '
             'for n_dim<50 the "gaussian" prior works fine and you should '
             'use it instead of the "gaussian_cached" prior'), UserWarning)
    # use a smaller logx_max as the point at which the logx at which the
    # scipy method (scipy.special.gammainc) fails is lower in lower
    # dimensions. logx_max=-10 will mean the gaussian_cached prior works
    # for all n_dim>2.
    if n_dim < 100:
        logx_max = kwargs.pop('logx_max', -10)
    elif n_dim < 250:
        logx_max = kwargs.pop('logx_max', -100)
    else:
        logx_max = kwargs.pop('logx_max', -200)
    if kwargs:
        raise TypeError('Unexpected **kwargs: {0}'.format(kwargs))
    save_name = (cache_dir + '/interp_gauss_prior_' + str(n_dim) + 'd_' +
                 str(prior_scale) + 'rmax_' + str(logx_min) + 'xmin_' +
                 str(logx_max) + 'xmax_' + str(interp_density) + 'id')
    try:
        interp_dict = iou.pickle_load(save_name)
    except (OSError, IOError):  # Python 2 and 3 compatable
        print(save_name)
        print('Interp file not found - try generating new data')
        r_max = mf.gaussian_r_given_logx(logx_max, prior_scale, n_dim)
        # Iteratively reduce r_min until its corresponding logx value is less
        # than logx_min. This process depends only on mpmath functions which
        # can handle arbitrarily small numbers.
        r_min = r_max
        while mf.gaussian_logx_given_r(r_min, prior_scale, n_dim) > logx_min:
            r_min /= 2.0
        r_array = np.linspace(r_min, r_max,
                              int((logx_max - logx_min) * interp_density))
        logx_array = mf.gaussian_logx_given_r(r_array, prior_scale, n_dim)
        interp_dict = {'interp_density': interp_density,
                       'logx_min': logx_min,
                       'n_dim': n_dim,
                       'prior_scale': prior_scale,
                       'r_array': r_array,
                       'logx_array': logx_array}
        if save_dict:
            iou.pickle_save(interp_dict, save_name)
    return interp_dict



#!/usr/bin/env python
"""
Mathematical functions.
"""


import numpy as np
import scipy
import scipy.stats
import scipy.special
import scipy.misc
import mpmath


def gaussian_r_given_logx(logx, sigma, n_dim):
    """
    Returns r coordinate corresponding to logx values for a Gaussian prior with
    the specificed standard deviation and dimension.

    This uses scipy.special.gammaincinv and requires exponentiating logx, so
    numerical errors occur with very low logx values.

    Parameters
    ----------
    logx: float or numpy array
        Logx coordinates at which to work out r.
    sigma: float
        Standard deviation of Gaussian.
    n_dim: int
        Number of dimensions.

    Returns
    -------
    float or numpy array
        Radial coordinates corresponding to input log X values.
    """
    exponent = scipy.special.gammaincinv(n_dim / 2., np.exp(logx))
    return np.sqrt(2 * exponent * sigma ** 2)


def gaussian_logx_given_r(r, sigma, n_dim):
    """
    Returns logx coordinate corresponding to r values for a Gaussian prior with
    the specificed standard deviation and dimension

    Uses mpmath package for arbitary precision.

    Parameters
    ----------
    r: float or numpy array
        Radial coordinates at which to evaluate logx.
    sigma: float
        Standard deviation of Gaussian.
    n_dim: int
        Number of dimensions.

    Returns
    -------
    logx: float or numpy array
        Logx coordinates corresponding to input radial coordinates.
    """
    exponent = 0.5 * (r / sigma) ** 2
    if isinstance(r, np.ndarray):  # needed to ensure output is numpy array
        logx = np.zeros(r.shape)
        for i, expo in enumerate(exponent):
            logx[i] = float(mpmath.log(mpmath.gammainc(n_dim / 2., a=0, b=expo,
                                                       regularized=True)))
        return logx
    else:
        return float(mpmath.log(mpmath.gammainc(n_dim / 2., a=0, b=exponent,
                                                regularized=True)))


def analytic_logx_terminate(settings):
    """
    Find logx_terminate analytically by assuming all likelihood at very low
    X approximately equals the maximum likelihood.
    This approximation breaks down in very high dimensions.

    Parameters
    ----------
    settings: PerfectNSSettings object

    Return
    ------
    float or None
    """
    # use r=0 rather than logx=-np.inf as the latter causes numerical problems
    logl_max = settings.logl_given_r(0)
    if settings.logz_analytic() is not None:
        return logx_terminate_bound(logl_max, settings.termination_fraction,
                                    settings.logz_analytic())
    else:
        return None


def logx_terminate_bound(logl_max, termination_fraction, logz_analytic):
    """
    Find a lower bound logx_terminate analytically by assuming all likelihood
    at very low X approximately equals the maximum likelihood. This
    approximation breaks down in very high dimensions and the true logx
    terminate required will be larger.

    We want:

    Z_term = termination_fraction * Z_analytic
           = int_0^Xterm L(X) dX
           approx= Xterm L_max,
    so
    logx_term = log(termination_fraction) + log(Z_analytic) - logl_max

    Parameters
    ----------
    logl_max: float
        Maximum loglikelihood.
    termination_frac: float
        Fraction of posterior mass remaining at termination.
    logz_analytic: float
        True value of log evidence.

    Returns
    -------
    float
    """
    return np.log(termination_fraction) + logz_analytic - logl_max


def sample_nsphere_shells_beta(r, n_dim, n_sample=None):
    """
    Given some 1d numpy array of radial coordinates r, return a numpy array
    of sample coordinates on the hyperspherical shells with each radial
    coordinate.

    Sample single parameters on n_dim-dimensional sphere independently, as
    as described in Section 3.2 of 'Sampling Errors in nested sampling
    parameter estimation' (Higson et al., 2018).

    NB if each parameter is sampled independently and combined into a vector,
    that vector will not have a magnitude r.

    The n_sample argument can be used to set the number of parameters for which
    samples are returned (by symmetry they all have the same distribution).
    This is useful for saving memory in high dimensions.

    This function is much quicker than sample_nsphere_shells_normal when n_dim
    is high and n_sample is low.

    Parameters
    ----------
    r: 1d numpy array
    n_dim: int
    n_sample: int or None, optional
        Number of parameters to include in output for each sample.

    Returns
    -------
    thetas: 2d numpy array
        Each row is a random sample from the shells defined by the
        corresponding input r coordinate.
    """
    assert isinstance(r, np.ndarray), 'input r must be a numpy array'
    if n_sample is None:
        n_sample = n_dim
    thetas = np.sqrt(np.random.beta(0.5, (n_dim - 1) * 0.5,
                                    size=(r.shape[0], n_sample)))
    # randomly select + or -
    thetas *= (-1) ** (np.random.randint(0, 2, size=thetas.shape))
    # multiply by r
    thetas *= r[:, None]
    return thetas


def sample_nsphere_shells_normal(r, n_dim, n_sample=None):
    """
    Given some 1d numpy array of radial coordinates r, return a numpy array
    of sample coordinates on the hyperspherical shells with each radial
    coordinate.

    This works by using the symmetry of the normal distribution to sample
    isotropically, then normalising each set of samples to lie on a spherical
    shell of the correct radius.

    The n_sample argument can be used to set the number of parameters for which
    samples are returned (by symmetry they all have the same distribution).
    This is useful for saving memory in high dimensions.

    Parameters
    ----------
    r: 1d numpy array
    n_dim: int
    n_sample: int or None, optional
        Number of parameters to include in output for each sample.

    Returns
    -------
    thetas: 2d numpy array
        Each row is a random sample from the shells defined by the
        corresponding input r coordinate.
    """
    if n_sample is None:
        n_sample = n_dim
    assert n_sample <= n_dim, 'so far only set up for nsample <= ndim'
    thetas = np.random.normal(size=(r.shape[0], n_dim))
    # calculate normalisation so sum_i(theta_i^2) = r^2 for each row
    norm = r / np.sqrt(np.sum(thetas ** 2, axis=1))
    # only return n_sample columns
    thetas = thetas[:, :n_sample]
    # normalise each column
    thetas *= norm[:, None]
    return thetas


def sample_nsphere_shells(r, n_dim, n_sample):
    """
    Wrapper calling either sample_nsphere_shells_normal or
    sample_nsphere_shells_beta depending on the dimension and
    number of samples required.

    See the docstrings of sample_nsphere_shells_normal and
    sample_nsphere_shells_beta for more information.
    """
    if n_dim >= 100 and n_sample <= 2:
        return sample_nsphere_shells_beta(r, n_dim, n_sample)
    else:
        return sample_nsphere_shells_normal(r, n_dim, n_sample)


def nsphere_r_given_logx(logx, r_max, n_dim):
    """
    Finds r coordinates given input logx values for a uniform prior within an
    n-dimensional sphere co-centred with a spherically symmetric likelihood.
    This will return an answer of the same type (float or numpy array) as the
    input {logx}.

    Parameters
    ----------
    logx: float or numpy array
    r_max: float
        Boundry of the spherically symmetric prior.
    n_dim: int
        Number of dimensions

    Returns
    r: float or numpy array
        Radial coordinates corresponding to logx.
    """
    r = np.exp(logx / n_dim) * r_max
    return r


def nsphere_logx_given_r(r, r_max, n_dim):
    """
    Finds logx assuming the prior is an n-dimensional sphere co-centred with a
    spherically symmetric likelihood.
    This will return an answer of the same type (float or numpy array) as the
    input {r}.

    Parameters
    ----------
    r: float or numpy array
    r_max: float
    n_dim: int

    Returns
    -------
    logx: float or numpy array.
        Logx coordinates corresponding to input radial coordinates.
    """
    logx = (np.log(r) - np.log(r_max)) * n_dim
    return logx


def nsphere_logvol(dim, radius=1.0):
    """
    Returns the natural log of the hypervolume of a unit nsphere of specified
    dimension. Useful for very high dimensions. Formulae is from
    https://en.wikipedia.org/wiki/N-sphere#Volume_and_surface_area

    Parameters
    ----------
    dim: int
    radius: float, optional

    Returns
    -------
    float
    """
    return ((np.log(radius) * dim) + (np.log(np.pi) * (dim / 2.0)) -
            (scipy.special.gammaln((dim / 2.0) + 1.0)))


def log_gaussian_given_r(r, sigma, n_dim):
    """
    Returns the natural log of a normalised, uncorrelated Gaussian likelihood
    with equal variance in all n_dim dimensions.

    Parameters
    ----------
    r: float or numpy array
    sigma: float
    n_dim: int

    Returns
    -------
    logl: float or numpy array
        Loglikelihood values corresponding to input radial coordinates.
    """
    logl = -0.5 * ((r ** 2) / (sigma ** 2))
    # normalise
    logl -= n_dim * np.log(sigma)
    logl -= np.log(2 * np.pi) * (n_dim / 2.0)
    return logl


def log_exp_power_given_r(r, sigma, n_dim, b=0.5):
    """
    Returns the natural log of an exponential power distribution.
    This equals a Gaussian distribution when b=1.

    Parameters
    ----------
    r: float or numpy array
    sigma: float
    n_dim: int
    b: float, optional

    Returns
    -------
    logl: float or numpy array
        Loglikelihood values corresponding to input radial coordinates.
    """
    logl = -0.5 * (((r ** 2) / (sigma ** 2)) ** b)
    # normalise
    logl += np.log(n_dim)
    logl += scipy.special.gammaln((n_dim) / 2.0)
    logl -= np.log(np.pi) * (n_dim / 2.0)
    logl -= n_dim * np.log(sigma)
    logl -= np.log(2) * (1 + 0.5 * b)
    logl -= scipy.special.gammaln(1. + ((n_dim) / (2 * b)))
    return logl


def r_given_log_exp_power(logl, sigma, n_dim, b=0.5):
    """
    Returns the natural log of an exponential power distribution.
    This equals a Gaussian distribution when b=1.

    Parameters
    ----------
    logl: float or numpy array
    sigma: float
    n_dim: int
    b: float, optional

    Returns
    -------
    r: float or numpy array
        Radial coordinates corresponding to input logl values.
    """
    # remove normalisation constant
    exponent = logl - np.log(n_dim)
    exponent -= scipy.special.gammaln((n_dim) / 2.0)
    exponent += np.log(np.pi) * (n_dim / 2.0)
    exponent += n_dim * np.log(sigma)
    exponent += np.log(2) * (1 + 0.5 * b)
    exponent += scipy.special.gammaln(1. + ((n_dim) / (2 * b)))
    # rearrange
    exponent = (-2. * exponent) ** (1. / b)
    logl = np.sqrt(exponent * (sigma ** 2))
    return logl


def r_given_log_gaussian(logl, sigma, n_dim):
    """
    Returns the radius of a given logl for a normalised,  uncorrelated Gaussian
    with equal variance in all n_dim dimensions.

    Parameters
    ----------
    logl: float or numpy array
    sigma: float
    n_dim: int

    Returns
    -------
    r: float or numpy array
        Radial coordinates corresponding to input logl values.
    """
    # remove normalisation constant
    exponent = logl + n_dim * np.log(sigma)
    exponent += np.log(2 * np.pi) * (n_dim / 2.0)
    # rearrange
    r = np.sqrt(-2 * exponent) * sigma
    return r


def log_cauchy_given_r(r, sigma, n_dim):
    """
    Returns the natural log of a normalised,  uncorrelated Cauchy distribution
    with 1 degree of freedom.

    Parameters
    ----------
    r: float or numpy array
    sigma: float
    n_dim: int

    Returns
    -------
    logl: float or numpy array
        Loglikelihood values corresponding to input radial coordinates.
    """
    logl = (-(1 + n_dim) / 2) * np.log(1 + ((r ** 2) / (sigma ** 2)))
    logl += scipy.special.gammaln((1.0 + n_dim) / 2.0)
    logl -= np.log(np.pi) * (n_dim + 1.0) / 2.0  # NB gamma(0.5) = sqrt(pi)
    logl += (-n_dim) * np.log(sigma)
    return logl


def r_given_log_cauchy(logl, sigma, n_dim):
    """
    Returns the radius for a given logl of a normalised,  uncorrelated
    Cauchy distribution with 1 degree of freedom.

    Parameters
    ----------
    logl: float or numpy array
    sigma: float
    n_dim: int

    Returns
    -------
    r: float or numpy array
        Radial coordinates corresponding to input logl values.
    """
    exponent = logl
    # remove normalisation constant
    exponent -= scipy.special.gammaln((1.0 + n_dim) / 2.0)
    exponent += np.log(np.pi) * (n_dim + 1.0) / 2.0  # NB gamma(0.5) = sqrt(pi)
    exponent -= (-n_dim) * np.log(sigma)
    # remove power
    exponent /= -(n_dim + 1.0) / 2.0
    # rearrange
    exponent = np.exp(exponent) - 1.0
    r_squared = exponent * (sigma ** 2)
    return np.sqrt(r_squared)



#!/usr/bin/env python
"""
perfectns setup module.

Based on https://github.com/pypa/sampleproject/blob/master/setup.py.
"""
import os
import setuptools


def get_package_dir():
    """Get the file path for dyPolyChord's directory."""
    return os.path.abspath(os.path.dirname(__file__))


def get_long_description():
    """Get PyPI long description from its own .rst file as PyPI does not render
    the README well."""
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, 'README.rst')) as readme_file:
        long_description = readme_file.read()
    return long_description


def get_version():
    """Get single-source __version__."""
    pkg_dir = get_package_dir()
    with open(os.path.join(pkg_dir, 'perfectns/_version.py')) as ver_file:
        string = ver_file.read()
    return string.strip().replace('__version__ = ', '').replace('\'', '')


setuptools.setup(name='perfectns',
                 version=get_version(),
                 description=('Dynamic and standard nested sampling '
                              'for spherically symmetric likelihoods and '
                              'priors.'),
                 long_description=get_long_description(),
                 long_description_content_type='text/x-rst',
                 url='https://github.com/ejhigson/perfectns',
                 author='Edward Higson',
                 author_email='e.higson@mrao.cam.ac.uk',
                 license='MIT',
                 keywords='nested-sampling dynamic-nested-sampling',
                 classifiers=[  # Optional
                     'Development Status :: 5 - Production/Stable',
                     'Intended Audience :: Science/Research',
                     'License :: OSI Approved :: MIT License',
                     'Programming Language :: Python :: 3',
                     'Programming Language :: Python :: 3.5',
                     'Programming Language :: Python :: 3.6',
                     'Topic :: Scientific/Engineering :: Astronomy',
                     'Topic :: Scientific/Engineering :: Physics',
                     'Topic :: Scientific/Engineering :: Visualization',
                     'Topic :: Scientific/Engineering :: Information Analysis',
                 ],
                 packages=['perfectns'],
                 install_requires=['numpy>=1.13',
                                   'scipy>=1.0.0',
                                   'pandas',
                                   'matplotlib>=2.1.0',
                                   'mpmath',
                                   'nestcheck>=0.1.2'],
                 test_suite='nose.collector',
                 extras_require={
                     'docs': ['sphinx', 'numpydoc', 'sphinx-rtd-theme',
                              'nbsphinx>=0.3.3']},
                 tests_require=['nose', 'coverage'],
                 project_urls={  # Optional
                     'Docs': 'http://perfectns.readthedocs.io/en/latest/'})



#!/usr/bin/env python
"""
Classes representing spherically symmetric priors.

Each prior class must contain a member function giving the radial
coordinate r = |theta| as a function of the log fraction of the prior volume
remaining and the dimension

    def r_given_logx(self, logx, n_dim):
        ...

The number of dimensions is not stored in this class but in the
PerfectNSSettings object to ensure it is the same for the likelihood and the
prior.

Prior classes may also optionally contain the inverse function

    def logx_given_r(self, r, n_dim):
        ...

(although this is not needed to generate nested sampling runs).
"""

import numpy as np
import scipy
import perfectns.maths_functions as mf
import perfectns.cached_gaussian_prior as cgp


class Uniform(object):

    """Spherically symmetric uniform prior."""

    def __init__(self, prior_scale):
        """Save the radius of the spherically symmetric uniform prior.

        prior_scale: float
            Radius of the uniform prior.
        """
        self.prior_scale = prior_scale

    def r_given_logx(self, logx, n_dim):
        """Maps input logx values to radial coordinates.

        Parameters
        ----------
        logx: float or numpy array
        n_dim: int

        Returns
        -------
        r: same type and size as logx
        """
        return mf.nsphere_r_given_logx(logx, self.prior_scale, n_dim)

    def logx_given_r(self, r, n_dim):
        """Maps input radial coordinates to logx values.

        Parameters
        ----------
        r: float or numpy array
        n_dim: int

        Returns
        -------
        logx: same type and size as r
        """
        return mf.nsphere_logx_given_r(r, self.prior_scale, n_dim)


class Gaussian(object):

    """Spherically symmetric Gaussian prior."""

    def __init__(self, prior_scale):
        """Save the input sigma for the prior. This is the (shared) standard
        deviation of each parameter.

        prior_scale: float
            Sigma for every parameter
        """
        self.prior_scale = prior_scale

    def r_given_logx(self, logx, n_dim):
        """Maps input logx values to radial coordinates.

        Parameters
        ----------
        logx: float or numpy array
        n_dim: int

        Returns
        -------
        r: same type and size as logx
        """
        return mf.gaussian_r_given_logx(logx, self.prior_scale, n_dim)

    def logx_given_r(self, r, n_dim):
        """Maps input radial coordinates to logx values.

        Parameters
        ----------
        r: float or numpy array
        n_dim: int

        Returns
        -------
        logx: same type and size as r
        """
        return mf.gaussian_logx_given_r(r, self.prior_scale, n_dim)


class GaussianCached(object):

    """
    Spherically symmetric uniform prior.
    The scipy inverse gamma function is not numerically stable so cache
    r_given_logx by using the mpmath logx_given_r and linearly
    interpolating.
    """
    interp_d = {'n_dim': None, 'prior_scale': None}

    def __init__(self, prior_scale, **kwargs):
        """Save input prior size, as well as parameters defining
        interpolation arrays and their caching.

        Parameters
        ----------
        prior_scale: float
            Sigma for every parameter.
        n_dim: int or None, optional
            Number of dimensions. Used to allow the prior to be calculated and
            cached before it is called.
        logx_min: float, optional
            See the cached_gaussian_prior.interp_r_logx_dict docstring for more
            information.
        save_dict: bool, optional
            See the cached_gaussian_prior.interp_r_logx_dict docstring for more
            information.
        cache_dir: str, optional
            See the cached_gaussian_prior.interp_r_logx_dict docstring for more
            information.
        interp_density: float, optional
            See the cached_gaussian_prior.interp_r_logx_dict docstring for more
            information.
        logx_max: float, optional
            See the cached_gaussian_prior.interp_r_logx_dict docstring for more
            information.
        """
        save_dict = kwargs.pop('save_dict', True)
        n_dim = kwargs.pop('n_dim', None)
        interp_density = kwargs.pop('interp_density', 10)
        logx_min = kwargs.pop('logx_min', -4500)
        cache_dir = kwargs.pop('cache_dir', 'cache')
        if kwargs:
            raise TypeError('Unexpected **kwargs: {0}'.format(kwargs))
        self.prior_scale = prior_scale
        self.save_dict = save_dict
        self.interp_density = interp_density
        self.logx_min = logx_min
        self.cache_dir = cache_dir
        # if n_dim is specified we can cache the interpolation now.
        # Otherwise wait until r_given_logx is called.
        if n_dim is not None:
            self.check_cache(n_dim)

    def r_given_logx(self, logx, n_dim):
        """Maps input logx values to radial coordinates.

        Parameters
        ----------
        logx: float or numpy array
        n_dim: int

        Returns
        -------
        r: same type and size as logx
        """
        self.check_cache(n_dim)
        if isinstance(logx, np.ndarray):
            r = np.zeros(logx.shape)
            ind = np.where(logx <= self.interp_d['logx_array'].max())[0]
            r[ind] = self.interp_f(logx[ind])
            ind = np.where(logx > self.interp_d['logx_array'].max())[0]
            r[ind] = mf.gaussian_r_given_logx(logx[ind], self.prior_scale,
                                              n_dim)
            assert np.count_nonzero(r) == r.shape[0], \
                'r contains zeros! r = ' + str(r)
            return r
        else:
            if logx <= self.interp_d['logx_array'].max():
                return self.interp_f(logx)
            else:
                return mf.gaussian_r_given_logx(logx, self.prior_scale,
                                                n_dim)

    def logx_given_r(self, r, n_dim):
        """Maps input radial coordinates to logx values.

        Parameters
        ----------
        r: float or numpy array
        n_dim: int

        Returns
        -------
        logx: same type and size as r
        """
        return mf.gaussian_logx_given_r(r, self.prior_scale, n_dim)

    def check_cache(self, n_dim):
        """
        Helper function which checks that the input dimension matches that of
        the cached interpolation function, and if needed recalculates it.

        Parameters
        ----------
        n_dim: int
            Number of dimensions
        """
        if (n_dim != self.interp_d['n_dim']) or (self.prior_scale !=
                                                 self.interp_d['prior_scale']):
            print('re-cache prior: input (n_dim, self.prior_scale) = (' +
                  str(n_dim) + ', ' + str(self.prior_scale) + ') =! cached ' +
                  '(n_dim, prior_scale) = (' + str(self.interp_d['n_dim']) +
                  ', ' + str(self.prior_scale) + ')')
            self.interp_d = cgp.interp_r_logx_dict(
                n_dim, self.prior_scale, save_dict=self.save_dict,
                cache_dir=self.cache_dir,
                logx_min=self.logx_min, interp_density=self.interp_density)
            self.interp_f = scipy.interpolate.interp1d(
                self.interp_d['logx_array'], self.interp_d['r_array'])



#!/usr/bin/env python
"""
Defines base class which holds settings for performing perfect nested sampling.
This includes likelihood and prior objects as well as the parameters
controlling how the calculation is performed - for example the number of live
points and whether or not dynamic nested sampling is to be used.
"""
# pylint: disable=no-member

import copy
import perfectns.priors as priors
import perfectns.likelihoods as likelihoods


class PerfectNSSettings(object):

    """
    Controls how Perfect Nested Sampling is performed.

    For more details of the standard nested sampling and dynamic nested
    sampling algorithms see:

    1) 'Dynamic nested sampling: an improved algorithm for nested  sampling
    parameter estimation and evidence calculation' (Higson et al.
    2017).
    2) 'Sampling errors in nested sampling parameter estimation' (Higson et al.
    2018)

    Parameters
    ----------
    n_dim: int
        Dimension of the likelihood and prior.
    prior: object
        A prior object. See priors.py for more details.
    likelihood: object
        A likelihood object. See likelihoods.py for more details.
    nlive_const: int
        The number of live points for standard nested sampling.
    termination_fraction: float
        Standard nested sampling runs will terminate when the posterior mass
        remaining (estimated from the live points) is less than
        termination_fraction times the posterior mass (evidence) in the dead
        points.
    dynamic_goal: float or None
        Determines is dynamic nested sampling is used, and if so how
        improvements in evidence and parameter estimation accuracy are
        weighted.
    n_samples_max: int or None
        Number of samples after which dynamic nested sampling will terminate.
        If this is None the number of samples to use is chosen using
        nlive_const.
    ninit: int
        Number of live points used in dynamic nested sampling for the initial
        exploratory standard nested sampling run.
    nbatch: int
        Number of threads dynamic nested sampling adds before each
        recalculation of points' importances to the calculation.
    dynamic_fraction: float
        Dynamic nested sampling adds more samples wherever points' importances
        are greater than dynamic_fraction times the largest importance.
    tuned_dynamic_p: bool
        Determines if dynamic nested sampling is tuned for a specific parameter
        estimation problem.
    """

    __isfrozen = False

    def __init__(self, **kwargs):
        """Initialise settings object and store settings

        Parameters
        ----------
        kwargs: dict, optional
            See the class docstring for a description of the allowed settings.
        """
        default_settings = {
            # likelihood and prior settings
            # -----------------------------
            'n_dim': 10,
            'prior': priors.Gaussian(10),
            'likelihood': likelihoods.Gaussian(1),
            # calculation settings
            # --------------------
            'nlive_const': 200,
            'dims_to_sample': 1,
            'termination_fraction': 0.001,
            # Dynamic nested sampling settings - only used if dynamic_goal is
            # not None.
            'dynamic_goal': None,
            'n_samples_max': None,
            'ninit': 20,
            'nbatch': 1,
            'dynamic_fraction': 0.9,
            'tuned_dynamic_p': False
        }
        for (setting_name, default_value) in default_settings.items():
            setattr(self, setting_name,
                    kwargs.pop(setting_name, default_value))
        if kwargs:
            raise TypeError('Unexpected **kwargs: {0}'.format(kwargs))
        # prevent more settings from being added later
        self.__isfrozen = True

    def __setattr__(self, key, value):
        """
        Freeze the class to prevent unexpected settings from being accidentally
        added.

        Parameters
        ----------
        key: str
            Name of attribute
        value: any type
            Value of attribute
        """
        if self.__isfrozen and not hasattr(self, key):
            raise TypeError('Frozen PerfectNSSettings instance given ' +
                            'unexpected attribute: %r' % key)
        object.__setattr__(self, key, value)

    # functions of the spherically symmetric prior

    def logx_given_r(self, r):
        """Maps input radial coordinates to logx values.

        Parameters
        ----------
        r: float or numpy array
        n_dim: int

        Returns
        -------
        logx: same type and size as r
        """
        return self.prior.logx_given_r(r, self.n_dim)

    def r_given_logx(self, logx):
        """Maps input logx values to radial coordinates.

        Parameters
        ----------
        logx: float or numpy array
        n_dim: int

        Returns
        -------
        r: same type and size as logx
        """
        return self.prior.r_given_logx(logx, self.n_dim)

    # functions of the spherically symmetric likelihood

    def logl_given_r(self, r):
        """Maps input radial coordinates to loglikelihood values.

        Parameters
        ----------
        r: float or numpy array
        n_dim: int

        Returns
        -------
        logl: same type and size as r
        """
        return self.likelihood.logl_given_r(r, self.n_dim)

    def r_given_logl(self, logl):
        """Maps input loglikelihood values to radial coordinates.

        Parameters
        ----------
        r: float or numpy array
        n_dim: int

        Returns
        -------
        logl: same type and size as r
        """
        return self.likelihood.r_given_logl(logl, self.n_dim)

    # functions of both the likelihood and the prior

    def logl_given_logx(self, logx):
        """Maps input logx values to loglikelihoods.

        Parameters
        ----------
        logx: float or numpy array
        n_dim: int

        Returns
        -------
        logl: same type and size as logx
        """
        return self.logl_given_r(self.r_given_logx(logx))

    def logx_given_logl(self, logl):
        """Maps input loglikelihoods to logx values.

        Parameters
        ----------
        logl: float or numpy array
        n_dim: int

        Returns
        -------
        logx: same type and size as logl
        """
        return self.logx_given_r(self.r_given_logl(logl))

    def logz_analytic(self):
        """
        If available gives an analytically calculated value for the log
        evidence for the likelihood and prior (useful for checking results).

        This functionality is stored in the likelihood object. If it has not
        been set up then an error message is printed and nothing is returned.

        Returns
        -------
        float or None
            True value of logz or None if it is not available.
        """
        try:
            return self.likelihood.logz_analytic(self.prior, self.n_dim)
        except (AttributeError, AssertionError):
            print('No logz_analytic set up for ' +
                  type(self.likelihood).__name__ +
                  " likelihoods and " + type(self.prior).__name__ + " priors.")
            return None

    def get_settings_dict(self):
        """
        Returns a dictionary containing settings information. The names and
        parameters of the likelihoods and priors are stored instead of the
        objects themselves so they can be saved with pickle.

        Returns
        -------
        settings_dict: dict
        """
        settings_dict = copy.deepcopy(self.__dict__)
        # Replace the likelihood and prior objects with information about them
        # in the form of strings and dicts so the output can be saved with
        # pickle.
        settings_dict['likelihood'] = type(self.likelihood).__name__
        settings_dict['likelihood_args'] = self.likelihood.__dict__
        settings_dict['prior'] = type(self.prior).__name__
        if type(self.prior).__name__ == 'GaussianCached':
            settings_dict['prior_args'] = {'prior_scale':
                                           self.prior.prior_scale}
        else:
            settings_dict['prior_args'] = self.prior.__dict__
        return settings_dict

    def save_name(self, include_dg=True, include_samples_max=False):
        """
        Make a standard save name format for a given settings configoration.

        Parameters
        ----------
        include_dg: bool, optional
            Whether or not to include the dynamic_goal setting in save_name.
        include_samples_max: bool, optional
            Whether or not to include the n_samples_max setting in save_name.

        Returns
        -------
        save_name: str
        """
        save_name = ''
        if include_dg:
            save_name += 'dg' + str(self.dynamic_goal) + '_'
        save_name += str(self.n_dim) + 'd'
        # Likelihood information
        save_name += '_' + type(self.likelihood).__name__
        like_params = self.likelihood.__dict__
        for param in sorted(like_params):
            save_name += '_' + str(like_params[param])
        # Prior information
        save_name += '_' + type(self.prior).__name__
        save_name += '_' + str(self.prior.prior_scale)
        # Nested sampling settings
        save_name += '_' + str(self.nlive_const) + 'nlive'
        save_name += '_' + str(self.termination_fraction) + 'term'
        if self.dims_to_sample != 1:
            save_name += '_' + str(self.dims_to_sample) + 'dimsamp'
        if self.dynamic_goal is not None or include_dg is False:
            save_name += '_' + str(self.ninit) + 'ninit'
            if self.nbatch != 1:
                save_name += '_' + str(self.nbatch) + 'nbatch'
        if self.n_samples_max is not None and self.dynamic_goal is not None:
            save_name += '_'
            if self.nlive_const is None or include_samples_max:
                # only when needed as otherwise get an annoying
                # long number in names
                save_name += str(self.n_samples_max)
            save_name += 'sampmax'
        if self.tuned_dynamic_p is True and self.dynamic_goal is not None:
            save_name += '_tuned'
        if self.dynamic_fraction != 0.9:
            save_name += '_' + str(self.dynamic_fraction) + 'df'
        save_name = save_name.replace('.', '_')
        save_name = save_name.replace('-', '_')
        return save_name
